/***************************************************************************
                          apaintBoundary.cpp  -  description
                             -------------------
    begin                : Tue Aug 3 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "apaintBoundary.h"
#include "apaintimage.h"
#include <qimage.h>

Boundary::Boundary(int wid)
{
  int i;
  vert_segs = new int[wid+1];
  for (i=0;  i<wid+1; i++)
    vert_segs[i] = -1;

  max_empty_segs = wid;

  empty_segs_n = new int[wid+4];
  empty_segs_c = new int[wid+4];
  empty_segs_l = new int[wid+4];
  num_empty_n = num_empty_c = num_empty_l = 0;
  boundList.setAutoDelete(true);
}

Boundary::~Boundary()
{
  boundList.clear();
  delete[] vert_segs;
  delete[] empty_segs_n;
  delete[] empty_segs_c;
  delete[] empty_segs_l;
}

                                 
/*  Function definitions  */

void Boundary::find_empty_segs(QImage  *mask,
                     int          scanline,
                     int          *empty_segs,
                     int         &num_empty)
{
  int    x, val, last=-1;
  int wid = mask->width();
  
  num_empty = 0;

  if ((scanline < 0) || (scanline >= mask->height()))
  {
    empty_segs[num_empty++] = 0;
    empty_segs[num_empty++] = INT_MAX;
    return; 
  }
  
  empty_segs[num_empty++] = 0;
  for (x=0; x < wid; x++)
  {
    if (PIX_ON(*mask, x, scanline))
      val = 1;
    else
      val = -1;
      
    if (last != val)
      empty_segs[num_empty++] = x;
    last = val;
  }
 
  if (last>0)
    empty_segs[num_empty++] = x;

  empty_segs[num_empty++] = INT_MAX;
}


void Boundary::make_seg (int x1, int y1, int x2, int y2, bool open)
{
  BoundSeg *seg = new BoundSeg;

  seg->x1 = x1;
  seg->y1 = y1;
  seg->x2 = x2;
  seg->y2 = y2;
  seg->open = open;

  boundList.append(seg);
}

void Boundary::process_horiz_seg (int x1, int y1, int x2, int y2, bool open)
{
  /*  This procedure accounts for any vertical segments that must be
      drawn to close in the horizontal segments.                     */

  if (vert_segs[x1] >= 0)
    {
      make_seg (x1, vert_segs[x1], x1, y1, !open);
      vert_segs[x1] = -1;
    }
  else
    vert_segs[x1] = y1;

  if (vert_segs[x2] >= 0)
    {
      make_seg (x2, vert_segs[x2], x2, y2, open);
      vert_segs[x2] = -1;
    }
  else
    vert_segs[x2] = y2;

  make_seg (x1, y1, x2, y2, open);
}


void Boundary::make_horiz_segs(int start, int end, int scanline, int *empty, int num_empty, int top)
{
  int empty_index;
  int e_s, e_e;    /* empty segment start and end values */

  for (empty_index = 0; empty_index < num_empty; empty_index += 2)
  {
      e_s = *empty++;
      e_e = *empty++;
      if (e_s <= start && e_e >= end)
        process_horiz_seg (start, scanline, end, scanline, top);
      else if ((e_s > start && e_s < end) || (e_e < end && e_e > start))
        process_horiz_seg (QMAX (e_s, start), scanline, QMIN (e_e, end), scanline, top);
  }
}

// mask = bitMask, px,py = translate position, numNoundSeg = returned number of segments
BoundSeg* Boundary::generate_boundary(QImage *mask, int px, int py, int &numBoundSeg)
{
  int  scanline;
  int  i;
  int *tmp_segs;
  BoundSeg *bseg;
  int start = 0;
  int end = mask->height();
  
  Boundary *bnd = new Boundary(mask->width());
  if (!bnd || !mask) return 0;
  
  /*  Find the empty segments for the previous and current scanlines  */
  bnd->find_empty_segs (mask, start - 1, bnd->empty_segs_l, bnd->num_empty_l);
  bnd->find_empty_segs (mask, start , bnd->empty_segs_c, bnd->num_empty_c);

  for (scanline = start; scanline < end; scanline++)
  {
      /*  find the empty segment list for the next scanline  */
      bnd->find_empty_segs (mask, scanline + 1, bnd->empty_segs_n, bnd->num_empty_n);
      /*  process the segments on the current scanline  */
      for (i = 1; i < bnd->num_empty_c - 1; i += 2)
      {
         bnd->make_horiz_segs (bnd->empty_segs_c[i], bnd->empty_segs_c[i+1], scanline, bnd->empty_segs_l, bnd->num_empty_l, 1);
         bnd->make_horiz_segs (bnd->empty_segs_c[i], bnd->empty_segs_c[i+1], scanline+1, bnd->empty_segs_n, bnd->num_empty_n, 0);
      }

      /*  get the next scanline of empty segments, swap others  */
      tmp_segs = bnd->empty_segs_l;
      bnd->empty_segs_l = bnd->empty_segs_c;
      bnd->num_empty_l = bnd->num_empty_c;
      bnd->empty_segs_c = bnd->empty_segs_n;
      bnd->num_empty_c = bnd->num_empty_n;
      bnd->empty_segs_n = tmp_segs;
  }
  
  numBoundSeg = (int)bnd->boundList.count();
  BoundSeg *bs = new BoundSeg[numBoundSeg];
  for (i=0; i< numBoundSeg; i++)
  {
    bseg = bnd->boundList.at(i);
    bseg->x1+=px;
    bseg->x2+=px;
    bseg->y1+=py;
    bseg->y2+=py;
    memcpy(&bs[i], bseg, sizeof(BoundSeg));
  }  
  delete bnd;
  return bs;
}

