/***************************************************************************
                          apaintBrushList.cpp  -  description
                             -------------------
    begin                : Thu Jul 15 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "apaint.h"
#include "apaintBrushList.h"
#include <stdlib.h>
#include "apaintconfig.h"
#include "apaintUtil.h"
#include "iconViewData.h"
#include "apaintdoc.h"
#include "apaintimageUtil.h"

#define PFX    "brush"

static ApaintBrushList *brushObject = 0;

ApaintBrushList *ApaintBrushList::getObject()
{
 if (!brushObject)
  brushObject = new ApaintBrushList();

 return brushObject;
}

void ApaintBrushList::deleteObject()
{
  if (brushObject)
    delete brushObject;
}

ApaintBrushList::ApaintBrushList()
{
  diffColorTab = true;
  isUsingBrush = false;
  absOffset = false;
  curMaskPath = "";
  brushDirOK = ApaintConfig::makeTmpDir(TMPDIR, PFX_APAINT, brushDir);
  iList = new IconList(1, 100, 100);
  iList->setIsBrushList(true);
  iList->sortIcons(DTimeSort);
  offsetX = offsetY = 0;
  for (int i=0; i<256; i++)
   transTable[i] = i;
  hide();
  connect(this, SIGNAL(hasBrushes(bool)), apaint, SLOT(hasBrushes(bool)));
}

ApaintBrushList::~ApaintBrushList()
{
  ApaintConfig::clearTmpDir(brushDir); 
}

bool ApaintBrushList::isValid()
{  
  if (!brushDirOK || !iList->iconNum())
   return false;
   
  return true;
}

QString ApaintBrushList::getMaskName(QString imageName)
{
  QString baseName = ApaintUtil::getBaseName(imageName);
  QString dirName = ApaintUtil::getDirName(imageName);

  return (QString("/%1/.%2").arg(dirName).arg(baseName));
}

void ApaintBrushList::addBrush(Layer *selLayer)
{
 char *dirName;

  if (!brushDirOK) return;
  
  dirName = (char *) malloc(300);
  strcpy(dirName, brushDir.ascii());
  strcat(dirName, PFX);
  strcat(dirName, "XXXXXX");
  if (mkstemp(dirName)<0)
   { free(dirName); return; }

  QString name = QString(dirName);
  QString nameMask = getMaskName(name);
  QImage maskSrc = selLayer->getMask();
  ApaintImage img;

  free(dirName);

  img.newImage(selLayer->imageBuf.width(), selLayer->imageBuf.height(), selLayer->imageBuf.depth());
  selLayer->imageBuf.copyImageDataTo(&img);
  if (img.type() == LOAD_CMAP)
   {
    int i = ApaintImageUtil::getLightestColor(selLayer->imageBuf.colorTable(), selLayer->imageBuf.numColors());
    img.fill(i);
   }
  else if (img.type() == LOAD_TRUE_COL)
   img.fill(qRgb(255,255,255));
  bitBlt(&img, 0, 0, &selLayer->imageBuf, 0, 0, selLayer->imageBuf.width(),selLayer->imageBuf.height(), false ,
         selLayer->usingMask(), &maskSrc, selLayer->getProtectColors(), selLayer->getTransColors());  

  img.save(name);
  QImage maskDst = QImage(img.width(), img.height(), 1, 2, QImage::LittleEndian);
  if (selLayer->usingMask())
   {    
    maskDst.fill(0);
    copy1BitImage(maskDst, 0, 0, maskSrc, 0, 0, maskSrc.width(), maskSrc.height(),
                   true, &selLayer->imageBuf, selLayer->getTransColors());
   }
  else
   maskDst.fill(255); 
  
  maskDst.save(nameMask, "PNG");  
  emit listIsChanged();
  emit hasBrushes(true);
}

void ApaintBrushList::loadBrush(int numCol, QRgb *ctab)
{   
  if (!brushDirOK) return;
  if (!iList->iconNum()) return;  
  getBrush(); // reload brush
  brushPos = QRect(0, 0, curBrush.width(), curBrush.height()); 
  diffColorTab = false;
  if (numCol)
  {
    if (curBrush.type() != LOAD_CMAP)
    {
      curBrush.setImageType(LOAD_CMAP, 0, 30, false, false);
      diffColorTab = true;
    }
    else
    {
      int i, ncol;
      QRgb *dtab = curBrush.colorTable();
      ncol = QMIN(numCol, curBrush.numColors());
      for (i=0; i<ncol; i++)
        if (ctab[transTable[i]] != dtab[i])
        {
          diffColorTab = true;
          return;
        }
    }
  }
}

void ApaintBrushList::removeBrush()
{
  if (!brushDirOK) return;
  if (!iList->iconNum()) return;
  if (iList->getNumIcons() == 1) emit hasBrushes(false);
  QString name = getMaskName(iList->getFullName(iList->selectedIcon()));
  // remove image
  if (iList->removeIconAndFile(iList->selectedIcon()))
    QFile::remove(name);   // remove mask 
  if (isValid())
  {
    ApaintImage b = getBrush();
    brushPos = QRect(brushPos.x(), brushPos.y(), b.width(), b.height());
  }
}

ApaintImage ApaintBrushList::getBrush()
{ 
 QString name = iList->getFullName(iList->selectedIcon());
 if ((name == "") || (!iList->iconNum()))
   curBrush = ApaintImage(); 
 else if (curBrush.imageName() != name)
   curBrush.load(name);

 return (curBrush);
}

QString ApaintBrushList::getBrushName()
{ 
 static QString name;

 name = iList->getFullName(iList->selectedIcon());
 return(name);
}

QImage ApaintBrushList::getBrushMask()
{
 QString name = getMaskName(iList->getFullName(iList->selectedIcon()));
 if ((name == "") || (!iList->iconNum())) return QImage();
 
 if (curMaskPath != name)
 {
   curMaskPath = name;
   curMask.load(name, "PNG");
   if (curMask.bitOrder() != QImage::LittleEndian)
    curMask = curMask.convertBitOrder(QImage::LittleEndian);   
 }
 return(curMask);
}

void ApaintBrushList::mergeColorTables(Layer *selLayer)
{
  if (!brushDirOK) return;
  if (selLayer->imageBuf.type() != LOAD_CMAP ) return;
  if (!selLayer) return;
  if (!diffColorTab) return;
  
  for (int i=0; i<256; i++) transTable[i] = i;
  ApaintImage brushImg = getBrush();
  if (brushImg.isNull()) return;
  
  if ((brushImg.numColors() + selLayer->imageBuf.numColors()) > 255)
  {
    brushImg.setColorTable(selLayer->imageBuf.colorTable(), selLayer->imageBuf.numColors());
  }
  else
  {
   selLayer->imageBuf.mergeColorTables(&brushImg, transTable);
   selLayer->imageBuf.copyImageDataTo(&selLayer->origBuf);
  }
  
  diffColorTab = false;
}

void ApaintBrushList::updateBrushList(int fhei, bool showDirs)
{  
  if (!brushDirOK) return;
  iList->setToNewDir(brushDir, fhei, showDirs);  
}

void ApaintBrushList::show()
{
  if (!brushDirOK || !iList->iconNum())
    showBrush = false;
  else
    showBrush = true;
}

void ApaintBrushList::mirrorHorizontal()
{
  if (!isValid()) return;
  QRect prevPos = brushPos;
  QString name = iList->getFullName(iList->selectedIcon());
  QString maskName = getMaskName(iList->getFullName(iList->selectedIcon()));
  getBrush();
  if (!curBrush.isNull())
  {
   curBrush.mirrorHorizontal();
   curBrush.save(name);
  }
  getBrushMask();
  if (!curMask.isNull())
  {
   curMask = curMask.mirror(true, false);
   curMask.save(maskName, "PNG");
  }  
  emit listIsChanged();
  brushPos = prevPos;
}

void ApaintBrushList::mirrorVertical()
{
  if (!isValid()) return;
  QRect prevPos = brushPos;
  QString name = iList->getFullName(iList->selectedIcon());
  QString maskName = getMaskName(iList->getFullName(iList->selectedIcon()));
  getBrush();
  if (!curBrush.isNull())
  {
   curBrush.mirrorVertical();
   curBrush.save(name);
  }
  getBrushMask();
  if (!curMask.isNull())
  {
   curMask = curMask.mirror(false, true);
   curMask.save(maskName, "PNG");
  }
  emit listIsChanged();
  brushPos = prevPos;
}

void ApaintBrushList::mirrorFlip()
{
  if (!isValid()) return;
  QRect prevPos;
  QString name = iList->getFullName(iList->selectedIcon());
  QString maskName = getMaskName(iList->getFullName(iList->selectedIcon()));
  getBrush();
  if (!curBrush.isNull())
  {
   curBrush.mirrorFlip(FALSE);
   curBrush.save(name);
   setPos(QRect(brushPos.x(), brushPos.y(), curBrush.width(), curBrush.height()));
   prevPos = brushPos;
  }
  getBrushMask();
  if (!curMask.isNull())
  {
   curMask = flipQImage(&curMask);
   curMask.save(maskName, "PNG");
  }
  emit listIsChanged();
  brushPos = prevPos;
}

void ApaintBrushList::mirrorExplode()
{
  if (!isValid()) return;
  QRect prevPos;
  QString name = iList->getFullName(iList->selectedIcon());
  QString maskName = getMaskName(iList->getFullName(iList->selectedIcon()));
  getBrush();
  if (!curBrush.isNull())
  {
   curBrush.mirrorExplode();
   curBrush.save(name);
   setPos(QRect(brushPos.x(), brushPos.y(), curBrush.width(), curBrush.height()));
   prevPos = brushPos;
  }
  getBrushMask();
  if (!curMask.isNull())
  {
   curMask = explodeQImage(&curMask);
   curMask.save(maskName, "PNG");
  }
  emit listIsChanged();
  brushPos = prevPos;
}

  
