/***************************************************************************
                          apaintimage.h  -  description
                             -------------------
    begin                : Mon May 5 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef APAINTIMAGE_H
#define APAINTIMAGE_H

#include <qobject.h>
#include <qrect.h>
#include <qptrlist.h>

extern "C"{
#include "../../llib/loadimg.h"
#include "../../plib/print.h"
}

#define MAX_IMAGE 65520 /* max width and height */
#define PIX_ON(src, sx, sy) (*((src).scanLine(sy)+((sx)>>3)) & (1 << ((sx) & 7)))
#define SET_PIX(src, sx, sy) (*((src).scanLine(sy)+((sx)>>3)) |= (1 << ((sx) & 7)))

#define FILTER_IKAT_X			0
#define FILTER_IKAT_Y			1
#define FILTER_IKAT_XY			2
#define FILTER_COLOR_SHORTER_INSIDE_X	3
#define FILTER_COLOR_SHORTER_INSIDE_Y	4
#define FILTER_COLOR_SHORTER_OUTSIDE_X	5
#define FILTER_COLOR_SHORTER_OUTSIDE_Y	6
#define FILTER_MIX_X			7
#define FILTER_MIX_Y			8
#define FILTER_MIX_XY			9
#define FILTER_NOISE_XY			10
#define FILTER_NOISE_X			11
#define FILTER_NOISE_Y			12
#define FILTER_TILT_RIGHT		13
#define FILTER_TILT_LEFT		14
#define FILTER_TILT_DOWN		15
#define FILTER_TILT_UP			16
#define FILTER_CLEAR_IMAGE		17
#define FILTER_CLEAR_IMAGE_X		18
#define FILTER_CLEAR_IMAGE_Y		19
#define FILTER_MOVE_LEFT		20
#define FILTER_MOVE_RIGHT		21
#define FILTER_MOVE_UP			22
#define FILTER_MOVE_DOWN		23
#define FILTER_SPLIT_X			24
#define FILTER_SPLIT_Y			25
#define FILTER_JOIN_X			26
#define FILTER_JOIN_Y			27
#define FILTER_ZIGZAG_X			28
#define FILTER_ZIGZAG_Y			29
#define FILTER_ZIGZAG_XY		30
#define FILTER_WAVE_X			31
#define FILTER_WAVE_Y			32
#define FILTER_WAVE_XY			33
#define FILTER_MIRROR_X			34
#define FILTER_MIRROR_Y			35
#define FILTER_CORAL_1			36
#define FILTER_CORAL_2			37
#define FILTER_CORAL_3			38
#define FILTER_CORAL_4			39
#define FILTER_CORAL_5			40
#define FILTER_CORAL_6			41
#define FILTER_SELF_AVOID_MOTIFS1	42
#define FILTER_SELF_AVOID_MOTIFS2	43
#define FILTER_SELF_AVOID_MOTIFS3	44
#define FILTER_SELF_AVOID_MOTIFS4	45
#define FILTER_SELF_AVOID_MOTIFS5	46
#define FILTER_SELF_AVOID_MOTIFS6	47
#define FILTER_SELF_AVOID_MOTIFS7	48
#define FILTER_SELF_AVOID_MOTIFS8	49
#define FILTER_SELF_AVOID_MOTIFS9	50
#define FILTER_RANDOM_MOTIFS1		51
#define FILTER_RANDOM_MOTIFS2		52
#define FILTER_RANDOM_MOTIFS3		53
#define FILTER_RANDOM_MOTIFS4		54
#define FILTER_RANDOM_MOTIFS5		55
#define FILTER_RANDOM_MOTIFS6		56
#define FILTER_RANDOM_MOTIFS7		57
#define FILTER_RANDOM_MOTIFS8		58
#define FILTER_RANDOM_MOTIFS9		59
#define FILTER_PLACE_MOTIFS1		60
#define FILTER_PLACE_MOTIFS2		61
#define FILTER_PLACE_MOTIFS3		62
#define FILTER_PLACE_MOTIFS4		63
#define FILTER_PLACE_MOTIFS5		64
#define FILTER_PLACE_MOTIFS6		65
#define FILTER_PLACE_MOTIFS7		66
#define FILTER_PLACE_MOTIFS8		67
#define FILTER_PLACE_MOTIFS9		68
#define FILTER_MARBLE_X			69
#define FILTER_MARBLE_Y			70
#define FILTER_MARBLE_XY		71
#define FILTER_CONTOUR			72
#define FILTER_GRADIENT_X		73
#define FILTER_GRADIENT_Y		74

/* contour directions; for selecting the direction of add and thicken operations */
#define CDIR_NONE	0
#define CDIR_LEFT	(1<<0)
#define CDIR_RIGHT	(1<<1)
#define CDIR_UP		(1<<2)
#define CDIR_DOWN	(1<<3)
#define CDIR_LEFT_UP	(1<<4)
#define CDIR_LEFT_DOWN	(1<<5)
#define CDIR_RIGHT_UP	(1<<6)
#define CDIR_RIGHT_DOWN	(1<<7)
#define CDIR_ALL	255

// used for filtering transformation
enum ResizeFilterType { FILTER_NONE, FILTER_BOX, FILTER_BICUBIC, FILTER_BILINEAR,
                        FILTER_BSPLINE, FILTER_CATMULLROM, FILTER_LANCZOS3, FILTER_SMOOTH2, FILTER_SMOOTH3, FILTER_SMOOTH4, };

class RemapArray {
public:
  RemapArray();
  ~RemapArray();

  void setLinear();
  void setVal(int pt, uchar val);
  uchar getVal(int pt);
private:
 uchar tab[256];
};  
  
class ApaintImageIO;
class ApaintImageUtil;  
    
class ApaintImage {
public: 
  ApaintImage();
  ApaintImage(const ApaintImage &other);
  ~ApaintImage();

  int bpl() {return data->imageBpl;}
  int type() {return data->imageType;}
  int width() {return data->imgWid;}
  int height() {return data->imgHei;}
  uchar **jumpTable() {return data->jumpTableData;}
  uchar *scanLine(int y) {return ((uchar *)(data->imageData+y*data->imageBpl));}
  QRgb *colorTable() {return(data->colorTableData);} 
  int numColors() {return(data->numColorsData);}
  int bytesPerLine() {return (data->imageBpl);}
  int numBytes() {return (data->imageBpl*data->imgHei);}
  void setColor(int i, QRgb col);
  QRgb color(int i);
  QRgb color(int i, int *realPos);
  void setNumColors(int i);
  bool setExactNumColors(int num);
  int depth();
  QRect rect();
  QRgb pixel(int x, int y);
  int pixelIndex(int x, int y);
  void setPixel(int x, int y, uint val);
  bool load(const QString &fname);
  bool save(QString &fname);
  bool isNull();
  void newImage(int wid, int hei, int depth);
  void newImageKeepName(int wid, int hei, int depth, ApaintImage *oldImg);
  void fill(int xr, int yr, int w, int h, uint col, bool useMask, QImage *mask);
  void fill(uint col);
  ApaintImage copy(int x, int y, int w, int h);
  ApaintImage copy();
//  ApaintImage copyOffsetX(int x, int y, int w, int h, int offx);
//  ApaintImage copyOffsetY(int x, int y, int w, int h, int offy);
  void horShift(int x);
  void vertShift(int y);
  ApaintImage& operator=(const ApaintImage& other);
  void setImageType(int imgType, QRgb *ctab, int numColors, int noise, bool dither);
  QString &imageName(){return data->imageName;}
  ApaintImage scale(int dw, int dh);
  ApaintImage rotate(double angle, ResizeFilterType ftyp, bool usingMask, QImage &mask);
  ApaintImage perspectiveXForm(int dstw, int dsth, double mx[3][3], bool usingMask, QImage &mask);
  ApaintImage xForm(const QWMatrix &mx, int dstWid, int dstHei, bool usingMask, QImage &mask);
  bool mergeColorTables(ApaintImage *src, int *transTable);
  void setRandomColors(int num);
  int xdpi(){ return data->imageXdpi;}
  int ydpi(){ return data->imageYdpi;}
  void setXdpi(int xdpi) {data->imageXdpi = xdpi;}
  void setYdpi(int ydpi) {data->imageYdpi = ydpi;}
  void setColorTable(QRgb *ctab, int num);
  ApaintImage resize(int wid, int hei, uint bgCol);
  void getUsedColors(bool *usedCol, int &num, bool useMask, QImage *mask);
  void remapData(RemapArray &transTable);
  void setExactColorTable(QRgb *ctab, int num);
  ApaintImage changeToBW(int treshHold);
  void mirrorHorizontal();
  void mirrorVertical();
  void mirrorFlip(bool noRotate);
  void mirrorExplode();
  void duplicate(int duplicateType);
  void scale2X();
  void scale3X();
  void shrink2X();
  void shrink3X();
  void shrink4X();
  void halve(int halveType);
  void shrink(int shrinkType);
  void multiplyX(int times);
  void multiplyY(int times);
  void copyImageDataTo(ApaintImage *img);
  void recalcJumpTable();
  void getHistogramData(int **histoNum, int **histoCol, bool useMask, QImage *mask);
  void removeUnusedColors();
  void changeColtabToBW(int treshHold);
  void mergeColors();
  int indexLastUsedColor();
  uchar *getCmap();
  void bwImagToCmap();
  void invertColors(int what, bool *protColors);
  void swapColors(int first, int second);
  void contrastStretchAP();
  void makeColorShades(int fgIndex, int bgIndex);
  void sortColors(int sortOrder, int selColor, int sortUp);

private:
  int initImageLoad();
  void setUpColorTable(bool reduce);
  void resampleColorTable(bool *validColor, int num);
  int bandImageLoad(int lines, char reverseLoad);
  int endImageLoad(int status, char reverseLoad);
  void freeImageData();  
  void saveColorTable();
  void refreshHistogram(bool useMask, QImage *mask);
  void cmap2TrueCol(uchar* dest, uchar *src);
  unsigned long getDistance(QRgb col1, QRgb col2);
  void setChangePosVal(int changecount, int pos, uint val);
  void flushChanges(bool *protColors, bool flushAll);
  bool allocDrawBuff(int size);
  void freeDrawBuff();
  void xFormHelper(QWMatrix &trueMat, int depth, uchar *dptr, int dbpl, int dHeight,
                   uchar *sptr, int sbpl, int sWidth, int sHeight, bool useMask, QImage &oldMask, QImage &mask);


  struct ImageSharedData
  {
   int refCount; 
   int imageBpl;
   int imageType;
   int imgWid, imgHei;
   int imageXdpi, imageYdpi;
   uchar *imageData;
   QRgb colorTableData[256];
   int numColorsData;
   int bandh, progress;
   uchar **jumpTableData;
   unsigned char cmap[256*3]; /* permanent cmap for colormap images */
   int histogramNum[256];
   int histogramCol[256];
   bool nullImage;
   QString imageName;
   unsigned long *changepos;
   uchar *changeval;
   int changecount;
  } *data;  

  friend class ApaintImageIO;
  friend class ApaintImageUtil;
  friend void bitBlt(ApaintImage *dst, int dx, int dy, ApaintImage *src, int sx, int sy, int sw, int sh,
                     bool sameImage, bool useMask, QImage *mask, bool *protectCol, bool *transCol);
  friend void bitBltAddCmap(ApaintImage *dst, int dx, int dy, ApaintImage *src, int sx, int sy, int sw, int sh,
                     bool sameImage, bool useMask, QImage *mask, bool *protectCol, bool *transCol, int groundColors, int layerColors);
  friend QDataStream &operator<<( QDataStream &s, ApaintImage &aimg );
  friend QDataStream &operator>>( QDataStream &s, ApaintImage &aimg );


};

class PrintDataProducer: public QObject{
  Q_OBJECT
public:
 PrintDataProducer(int pageWidth, int pageHeight, int bandHei, char imageType, bool landscape);
 virtual ~PrintDataProducer();

 uchar *curDataPtr();
 bool endOfData(){ return eod;}
 int pageWidth() {return mPageWidth;}
 int pageHeight() {return mPageHeight;}
 int bpl() {return mbpl;}
 int curLine() {return mcurLine;}
 int lastLine() {return toLine;}
 int bandHei() {return mbandHei;}
 char printType() {return mimgType;}
 int getPrtStatus() {return mprtStatus;}
 void setPrtStatus(int status) {mprtStatus = status;}
 void setFilename(QString fname) {mFilename = fname;}
 QString getFilename() {return mFilename;}
 void setSaveToFile(bool enable) {mSaveToFile = enable;}
 bool saveToFile() {return mSaveToFile;}
 int depth() {return mdepth;}
 
 virtual uchar *getColorMap();
 virtual QRgb *getColorTable();
 virtual int getNumColors();
 virtual void nextBand();

signals:
  void setProgress(PrintDataProducer *pdp);
  
protected:

 void initData();
 
 uchar *data;
 bool eod, mlandscape;
 int mPageWidth, mPageHeight, mbandHei;
 int mdepth, mbpp, mbpl, mcurLine, toLine;
 int mprtStatus;
 char mimgType;
 QString mFilename;
 bool mSaveToFile;
};

class PrintDataProducerImage: public PrintDataProducer {
public:
 PrintDataProducerImage(ApaintImage &img, int pageWidth, int pageHeight, int imgWidth, int imgHeight, int bandHei, char imageType, bool landscape);
 virtual ~PrintDataProducerImage();

  virtual void nextBand();
  virtual uchar *getColorMap();
  virtual QRgb *getColorTable();
  virtual int getNumColors();
  
private:
  ApaintImage mimg;
  int mImgWidth, mImgHeight; 
};


class ApaintImageIO {

private:
  ApaintImageIO();
  
public:  
	~ApaintImageIO();

  static ApaintImageIO *getObject();
  const ApaintImage& image() const {return aimg;}  
  bool loadFile(char *filename);
  bool saveFile(char *filename,  ApaintImage *img);
  bool printFile(int printerID, QPtrList<PrintDataProducer> &pdpList, int numCopies, int xdpi, int ydpi, int xmargin, int ymargin);
  int getNumPrinters() {return numPrinters;}
  printerMode *getPrinterModes(int i, int *num);
  int getPrinterMode(int prt, int i);
  printerDensity *getPrtDensityX(int i, int *num);
  int getDensityX(int prt, int i);
  printerDensity *getPrtDensityY(int i, int *num);
  int getDensityY(int prt, int i);
  char *getPrinterName(int i);
  int getNonPrintableLeft(int prt);
  int getNonPrintableTop(int prt);
  void selectPrinter(int printerID);
  void setPrintCapType(QString pcap);
  
  
private:
  void setupPrinterData();
  char const *getImageInfo(char *filename, int *x, int *y, unsigned char *loaderID, int *coreFile);
  int loadImage(char *filename, int desiredLoadType, unsigned char *loaderID);
  void processBand();
  void processSaveBand();
  void processPrintBand(QPtrList<PrintDataProducer> &pdpList);
  bool selectSaverPrinter(char *findPrinter);
  bool getFileInfo(char *filename, char *loaderID, int *changeImg, int imgType);
  
  typedef struct
  {
    printerMode prtModes[5];
    printerDensity prtDensityX[20];
    printerDensity prtDensityY[20];
    int numModes, numDensitX, numDensitY, nonprintableTop, nonprintableBottom, nonprintableLeft, dpivariable;
    char *printerName, *profileName;
  } PrinterInfo;
  
  PrinterInfo *pInfo;
  int numPrinters;
  ApaintImage aimg;
  int xmargin, ymargin;
  bool complexImage;
  int loadStatus, prtStatus; /* the status of loading */
  char loaderInfo[500]; /* tryload fills this in */
  char reverseLoad;
  int  printType;
  int numLoaders;
  QString pcapType;
};

QString getDirName(QString str);
QString getPfx(char *name);
bool brushFilename(QString &name);
bool fixPfx(QString &name);
void chgPfx(QString &name, QString pfx);

void bitBlt(ApaintImage *dst, int dx, int dy, ApaintImage *src, int sx, int sy, int sw, int sh,
           bool sameImage, bool useMask, QImage *mask, bool *protectCol, bool *transCol);

void bitBltAddCmap(ApaintImage *dst, int dx, int dy, ApaintImage *src, int sx, int sy, int sw, int sh,
           bool sameImage, bool useMask, QImage *mask, bool *protectCol, bool *transCol);

QImage flipQImage(QImage *src);
QImage explodeQImage(QImage *src);
QImage duplicateQImage(int duplicateType, QImage *src);
QImage scale2XQImage(QImage *src);
QImage scale3XQImage(QImage *src);
QImage shrink2XQImage(QImage *src);
QImage shrink3XQImage(QImage *src);
QImage shrink4XQImage(QImage *src);
QImage halveQImage(int halveType, QImage *src);
QImage shrinkQImage(int shrinkType, QImage *src);
QImage multiplyXQImage(int times, QImage *src);
QImage multiplyYQImage(int times, QImage *src);
QImage cutCorners(QImage& src, QRect &copyRect);
void copy1BitImage(QImage& dimg, int dx, int dy, QImage& simg, int sx, int sy, int sw, int sh,
                   bool useTransImage, ApaintImage *timg, bool *transCol);

QImage setMaskWithColor(ApaintImage& simg, uint bgCol);
QImage scaleQImage(int sx, int sy, QImage &src, QRect totalSize, QRect finalSize, int *srcZoomX, int *srcZoomY, int &dx, int &dy, int &dstw, int &dsth);
QImage invertQImage(QImage &src);

void flipImageX(uchar *data, int type, int wid, int hei, int bpl);
void flipImageY(uchar *data, int hei, int bpl);
void flipImageXY(bool noRotate, uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl);
void explodeImage(uchar *data, int type, int wid, int hei, int bpl);
void duplicateImage(int duplicateType, uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl);
void scale2XImage(uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void scale3XImage(uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void shrink2XImage(uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void shrink3XImage(uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void shrink4XImage(uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void halveImage(int halveType, uchar *srcData, uchar *dstData, int shei, int sbpl, int dbpl);
void shrinkImage(int shrinkType, uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void multiplyXImage(int times, uchar *srcData, uchar *dstData, int type, int swid, int shei, int sbpl, int dbpl);
void multiplyYImage(int times, uchar *srcData, uchar *dstData, int shei, int sbpl, int dbpl);

QDataStream &operator<<( QDataStream &s, ApaintImage &aimg );
QDataStream &operator>>( QDataStream &s, ApaintImage &aimg );

#endif
