/***************************************************************************
                          icondataview.cpp  -  description
                             -------------------
    begin                : Mon Jan 5 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <qcolor.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <sys/types.h>
#include <utime.h>

extern char *getCommonPNG(const char *pixName);

#include "iconViewData.h"
#include "apaintconfig.h"

#define ICON_CACHE_DAYS   7
/****************************ICON DATA**************************************/

IconData::IconData()
{
  iconPos = QRect();
  textPos = QRect();
  init();  
}

IconData::IconData(QRect iconPosp, QRect textPosp)
{
  iconPos = iconPosp;
  textPos = textPosp;
  init();
}

void IconData::init()
{  
  imgInfo = 0;
  type = None;
}

IconData::~IconData()
{
}

void IconData::setIconRect(int x, int y, int w, int h)
{
  iconPos.setRect(x, y, w, h);
}

void IconData::setTextRect(int x, int y, int w, int h)
{
  textPos.setRect(x, y, w, h);
}

void IconData::setImageInfo(imageInfo *info)
{
  imgInfo = info;
  type = ImageType;
}

void IconData::setDirInfo(dirBrowse *setDirB)
{
 dirB = setDirB;
 type = DirType;
}

bool IconData::isIconLoaded()
{
 if (type == ImageType)
  if (imgInfo->iconData)
   return true;
 
 return false;
}

void IconData::fillImgInfo()
{
  int browseImageBPL;
  if (!imgInfo) return;

  if (!isIconLoaded())
   loadImageMakeIcon(imgInfo, aCol, &browseImageBPL);
}

void IconData::renderIcon(unsigned char *dst, int bpp, int bpl)
{
 if (type == ImageType && imgInfo)
  {
    fillImgInfo();
    renderBrowseIconImage(imgInfo->iconData, imgInfo->compressedIconLen,
                          imgInfo->iconType, imgInfo->iconX, imgInfo->iconY, imgInfo->iconCmap,
                          iconx-2, 1, iconx, (ApaintConfig::getBrowseImageDisplay() & IMAGE_ICON_REPEAT) ? 0 : 1, bpp, dst, bpl);
  }
}

void IconData::setIconSize(int x, int y)
{
  iconx = x;
  icony = y;
}

QString IconData::getName()
{
 if (type == DirType)
  return QString::fromLocal8Bit(dirB->name);
 else if (type == ImageType)
  return QString::fromLocal8Bit(imgInfo->Fname); 
 else
  return QString();
}

QString IconData::getSize()
{
 if (type != ImageType) return QString();
 if (!imgInfo) return QString();
 QString str;
 off_t fsize = imgInfo->fileSize;
 
 if (fsize<1000)
  str = QString("%1").arg(fsize);
 else if (fsize<1000000)
  str.sprintf("%ld.%03ld ", fsize/1000, fsize%1000);
 else if (fsize<1000000000)
  str.sprintf("%ld.%03ld.%03ld", fsize/1000000, (fsize/1000)%1000, fsize%1000);
 else
  str.sprintf("%ld.%03ld.%03ld.%03ld", fsize/1000000000, (fsize/1000000)%1000, (fsize/1000)%1000, fsize%1000);

 str.append(" "+ApaintConfig::getText(BYTES_MSG));
 return str;
}

QString IconData::getInfo(bool retTrueColor, bool loaderID)
{
 if (type != ImageType) return QString();
 if (!imgInfo) return QString();
 QString str, ncol;

 if (imgInfo->imageType==LOAD_CMAP)
  ncol = QString("%1 %2").arg(getNumColors()).arg(ApaintConfig::getText(COLORS_MSG));
 else
  {
   ncol = "";
   if (imgInfo->imageType==LOAD_BW)
    ncol = ApaintConfig::getText(139);
   if (imgInfo->imageType==LOAD_RGBA)
    ncol = ApaintConfig::getText(636);
  }
      
 if (!loaderID)
   str = QString("%1x%2 %3").arg(imgInfo->imageX)
                            .arg(imgInfo->imageY)
                            .arg(ncol);
 else
   str = QString("%1x%2 %3 %4").arg(imgInfo->imageX)
                             .arg(imgInfo->imageY)
                             .arg(returnLoaderString(imgInfo->loaderID))
                             .arg(ncol);
 return str; 
}

QString IconData::getDTime()
{
  struct tm *tmTime;
  char infoStr[50];
  
  if (type != ImageType) return QString();
  if (!imgInfo) return QString();  
  
  tmTime =localtime(&imgInfo->modifiedTime);
  strftime(infoStr, 40, "%d.%b.%Y %H:%M", tmTime);
  return QString::fromLocal8Bit(infoStr);
}

int IconData::getXDpi()
{
  if (type != ImageType) return 100;
  if (!imgInfo) return 100;

  return imgInfo->xdpi;
}

int IconData::getYDpi()
{
  if (type != ImageType) return 100;
  if (!imgInfo) return 100;

  return imgInfo->ydpi;
}

int IconData::getImageX()
{
  return imgInfo->imageX;
}

int IconData::getImageY()
{
  return imgInfo->imageY;
}

int IconData::getNumColors()
{
  if (type != ImageType) return 0;
  if (!imgInfo) return 0;

  /* special case for 256 color palette image, since we only have one byte for number of colors */  
  if ((imgInfo->imageType==LOAD_CMAP) && (!imgInfo->usedColors))
   return(256);

  return(imgInfo->usedColors);
}

int IconData::getNumDirs()
{
 if (type == DirType)
  return(dirB->numDirs);
 else
  return(0);
}

int IconData::getNumFiles()
{
 if (type == DirType)
  return(dirB->numFiles);
 else
  return(0);
}

void IconData::getNumDirsFiles()
{
 char filen[500];

 if (type != DirType)
  return;

 /* do it only if view did not already enable directory numbers */
 if (ApaintConfig::getBrowseImageDisplay() & IMAGE_ICON_DIR_INFO)
  return;
 sprintf(filen, "%s/%s", getImageBrowsePath(), dirB->name);
 countDirFiles(filen, &dirB->numDirs, &dirB->numFiles);
}
                                           
/********************ICON LIST****************/

IconList::IconList(int numOfCol, int iSizeX, int iSizeY)
{
 selIcon = 0;
 numIcons = 0;
 numImages = 0;
 numDirs = 0;
 iList = 0;
 numCol = numOfCol;
 numRow = 0;
 allIconsInfo = 0;
 iconX =  iSizeX;
 iconY =  iSizeY;
 textY = 0;
 dirLoaded = false;
 iconDataMap.setAutoDelete(false);
 aCol = (aColor*) malloc(sizeof(aColor)*256);
 clearArea = QRect();
 folderIcon = new QPixmap(getCommonPNG("folder-32"));
 selectedSort = NameSort;
 *getIconCacheDays() = ApaintConfig::getClearIconCacheDays();
 initBrowseIO();
 usingFilter = false;
 includeSubDir = false;
 setImageBrowseIncludeSubDir(includeSubDir);
 for (int i=0; i<filterNum; i++)
  {
   ftyp[i] = None;
   fop[i] = More;
   fvalue[i] = 0;
  }
}

IconList::~IconList()
{
  if (iList!=0)
   delete[] iList;
  free(aCol);
}

void IconList::setIsBrushList(bool enable)
{
  swIsBrushList = enable;
}

bool IconList::renameIconAndFile(int num, QString fname)
{
 if (num<0 || num>=numIcons) return false;
 if (num<numDirs) return false;
 if (getIconData(num)->isDir()) return false;
 if (fname == "") return false;

 IconData *ic = getIconData(num);
 imageInfo *iI = ic->getImageInfo();

 QDir dir(iI->subDirID ? QString("%1%2").arg(imageDir).arg(QString::fromLocal8Bit(getImageSubDir(iI->subDirID))) : QString::fromLocal8Bit(getImageBrowsePath()), false);
 if (!dir.rename(getIconData(num)->getName(), fname))
  return false;

 utime(getIconData(num)->getName(), NULL); /* change the modified date of the file to current time, so icon will be correct */
 imageInfo *renImage = getIconData(num)->getImageInfo();
 free(renImage->Fname);
 renImage->Fname = strdup((char *)fname.local8Bit().data());

 return true;
}

// I hope it's faster than do setNewDir(sameDir)
bool IconList::removeIconAndFile(int num)
{  
  if (num<0 || num>=numIcons) return false;
  if (num<numDirs) return false;
  if (getIconData(num)->isDir()) return false;

 // remove the actual file
 if (!QFile::remove(getFullName(num)))
  return false;
  
 // delete from iolib list :imageInfo[]
 int icNum = iconData2FilterNum(num - numDirs);
 if (icNum == -1)
  return false;
 imageInfo *delImage = iconDataMap.at(icNum);
 imageInfo *lastImage = &allIconsInfo[numImages-1];
 if (delImage != lastImage)
   swapImageIcon(delImage, lastImage);
 freeImageI(lastImage);
 
 // set num data
 numImages--;
 numIcons = numImages+numDirs;
 if (selIcon >= numIcons)
  selIcon = numIcons-1;
  
 // reset my internal list: iList[]
 if (iList!=0)
 {
   delete[] iList;
   iList = 0;
 }
 iList = new IconData[numIcons];
 if (!iList)
 {
    numIcons = 0;
    dirLoaded = false;
    qDebug("IconList::removeIconAndFile: ptr NULL!\n");
    return false;
 }
 useFilter(usingFilter);
 fillMapTable();
 setIsBrushList(swIsBrushList);

 dirLoaded = true;
 // setSize and refresh must be called after this function
 return true;
}

//get file pos from iconDataMap with filterOK param checked.
int IconList::iconData2FilterNum(int num)
{
 int rnum = 0;

 for (unsigned int i = 0; i < iconDataMap.count(); i++)
  {
   if (iconDataMap.at(i)->filterOK)
    {
     if (rnum == num)
      return i;
     else
      rnum++;
    }
  }
 return(-1);
}

void  IconList::fillMapTable()
{
  iconDataMap.clear();
  for (int i=0; i<numImages; i++)
    iconDataMap.append(&allIconsInfo[i]);
  sortImages();
}

bool IconList::setToNewDir(QString dir, int thei, bool showDirs)
{
 selIcon = 0;
 textHeight = thei;
 //ICONS LOAD STUFF...
 //numIcons = number of icons in current dir.
 imageDir = dir;
 if (!imageDir.endsWith("/"))
   imageDir.append("/");

 if (dir.right(1) == "/")
  dir = dir.left(dir.length()-1);

 setBrowsePath((char *)dir.local8Bit().data());
 loadImageIcons((char *)dir.local8Bit().data(), &allIconsInfo, &numImages, &dirI, aCol, ApaintConfig::getBrowseImageDisplay());
 dirB = dirI->dirB;
 numDirs = dirI->numDir;

 if (!showDirs) numDirs = 0;
 
 numIcons = numImages+numDirs;  
 if (iList!=0)
 {
   delete[] iList;
   iList = 0; 
 }
 iList = new IconData[numIcons];
 if (!iList)
 {
   numIcons = 0;
   dirLoaded = false;
   qDebug("IconList::removeIconAndFile: ptr NULL!\n");
   return false;
 }
 useFilter(usingFilter);
 fillMapTable();
 setIsBrushList(swIsBrushList);

 dirLoaded = true;
 return true;
}

void IconList::setSize(int ncol, int sx, int sy)
{
  if (!dirLoaded) return;
  if (!numIcons)
   {
    iconX = iconY = numRow = 0;
    return;
   }

  int i, sc, iconNum, nc, nr;
  imageInfo *iinfo;

  numCol = QMIN(ncol, numIcons);
  numRow = numIcons / ncol;
  if (numIcons%ncol>0) numRow++;
  iconX = sx;
  iconY = sy;

  sc = 0;
  if (ApaintConfig::getBrowseImageDisplay() & IMAGE_ICON_FILENAME) sc++;
  if (ApaintConfig::getBrowseImageDisplay() & IMAGE_ICON_IMAGESIZE) sc++;
  if (ApaintConfig::getBrowseImageDisplay() & IMAGE_ICON_FILESIZE) sc++;
  if (ApaintConfig::getBrowseImageDisplay() & IMAGE_ICON_FILEDATE) sc++;
  textY = QMAX(1,textHeight * (swIsBrushList ? 1 : sc));

  for (i=0; i<numDirs; i++)
  {
    nc = i%numCol;
    nr = i/numCol;
    iList[i].setDirInfo(&dirB[i]);
    iList[i].setIconRect(nc*(iconX), nr*(iconY+textY), iconX, iconY);
    iList[i].setIconSize(iconX, iconY);
    if (textY)
      iList[i].setTextRect(nc*(iconX), nr*(iconY+textY)+iconY, iconX, textY);
  }
  iconNum = numDirs;
  for (i=0; i<numImages; i++)
  {
    iinfo = iconDataMap.at(i);
    if (iinfo->filterOK)
     {
      nc = iconNum%numCol;
      nr = iconNum/numCol;
      iList[iconNum].setImageInfo(iinfo);
      iList[iconNum].setIconRect(nc*(iconX), nr*(iconY+textY), iconX, iconY);
      iList[iconNum].setIconSize(iconX, iconY);
      if (textY)
        iList[iconNum].setTextRect(nc*(iconX), nr*(iconY+textY)+iconY, iconX, textY);
      iconNum++;
     }
 }
 int rest = numIcons%ncol;
 if (rest>0)
  clearArea = QRect(rest*iconX, (numRow-1)*(iconY+textY), iconX*(ncol-rest), textY+iconY);
 else
  clearArea = QRect();
}

IconData* IconList::getIconData(int num)
{
  if (!dirLoaded) return 0;
  if (num<0 || num >= numIcons) return 0;

  return &iList[num];
}

QString IconList::getFullName(int num)
{
 IconData *ic = getIconData(num);
 if (!ic) return "";
 imageInfo *iI = ic->getImageInfo();

 if (iI && iI->subDirID)
  return QString("%1%2/%3").arg(imageDir).arg(QString::fromLocal8Bit(getImageSubDir(iI->subDirID))).arg(ic->getName());
 else
  return QString("%1%2").arg(imageDir).arg(ic->getName());
}

int IconList::getNewPosByFirstCharacter(int first)
{
 int pos, start = 0;
 int found = FALSE;
 const char *name;

 if (numIcons<2)
  return(0);

 while (!found && (start<numIcons))
  {
   start++;
   pos = (selIcon+start)%numIcons;
   if (pos<numDirs)
    {
     if (dirB[pos].name[0]==first)
      found=TRUE;
    }
   else
    {
     IconData *ic = getIconData(pos);
     imageInfo *iI = ic->getImageInfo();

     name = iI->Fname;
     if (name[0]==first)
      found=TRUE;
    }
  }
 return(found ? pos : -1);
}

void IconList::setSelectedIcon(int num)
{
  if (num<0 || num>=numIcons) return;
  selIcon = num;
}

void IconList::setSelectedIcon(int x, int y)
{
  int n = y*numCol+x;
  
  if (n>numIcons)
   selIcon = numIcons-1;
  else if (n<0)
   selIcon = 0;
  else
   selIcon = n;
}

bool IconList::validPos(int x, int y)
{
 int n = y*numCol+x;
 if (n<0 || n>=numIcons)
  return false;

 return true;
}

int IconList::getIconFromPos(int x, int y)
{
  for (int i=0; i<iconNum(); i++)
  {
    IconData *ic = getIconData(i);
    if (ic->getIconPos().contains(x, y) || ic->getTextPos().contains(x, y))
     return i;
  }

  return -1;
}

void IconList::sortImages()
{
  iconDataMap.setSortType(selectedSort);
  iconDataMap.sort();
}

void IconList::sortIcons(SortType styp)
{
 selectedSort = styp;
 sortImages();
}


void IconList::calcFilter()
{
  int filterIconsNum=0;
  int cmpVal = 0, cmpVal2 = 0;
  imageInfo *iinfo;
  bool satisfied;
  time_t ctime;
  QString cmpStr;

  time(&ctime);
  for (int i=0; i<numImages; i++)
  {
    satisfied = true;
    iinfo = &allIconsInfo[i];
    for (int f=0; (f<filterNum && satisfied); f++)
    {
      if (ftyp[f] == None) continue;
      switch(ftyp[f])
       {
        case SizeX:
          cmpVal = iinfo->imageX;
          break;
        case SizeY:
          cmpVal = iinfo->imageY;
          break;
        case SizeXY:
          cmpVal =  iinfo->imageX;
          cmpVal2 = iinfo->imageY;
          break;
        case DaysOld:
          cmpVal = (ctime-iinfo->modifiedTime)/(60*60*24);
          break;
        case Colors:
          cmpVal = iinfo->usedColors ? iinfo->usedColors : 256;
          break;

        case None:
         break;  
       }
      switch(fop[f])
       {
        case More:
          satisfied = cmpVal > fvalue[f];
          break;
        case Less:
          satisfied = cmpVal < fvalue[f];
          break;
        case Equal:
          satisfied = cmpVal == fvalue[f];
          break;
       }    
     if (satisfied && cmpVal2)
      switch(fop[f])
       {
        case More:
          satisfied = cmpVal2 > fvalue[f];
          break;
        case Less:
          satisfied = cmpVal2 < fvalue[f];
          break;
        case Equal:
          satisfied = cmpVal2 == fvalue[f];
          break;
       }    
   }

    if (fStr.length()>0 && satisfied)
     {
      cmpStr = QString::fromLocal8Bit((iinfo->Fname)).lower();
      satisfied =cmpStr.contains(fStr, false);
     }
    iinfo->filterOK = satisfied;
    if (satisfied)
     filterIconsNum++;
  }
  numIcons = numDirs+filterIconsNum;
}

void IconList::setFilter(int nFilter, FilterType typ, FilterOp op, int value)
{
  if (nFilter<0 || nFilter>filterNum) return;

  ftyp[nFilter] = typ;
  fop[nFilter] = op;
  fvalue[nFilter] = value;
  calcFilter();    
}

void IconList::setFilterText(QString cmpStr)
{
  fStr = cmpStr;
  fStr.lower();
  calcFilter();
}

void IconList::clearFilter()
{
  for (int i=0; i<numImages; i++)
   allIconsInfo[i].filterOK = 1; 
  numIcons = numDirs+numImages;
}

void IconList::useFilter(bool enable)
{
  usingFilter = enable;
  if (enable)
   calcFilter();
  else
   clearFilter();
}

QString IconList::getTotalFileSize()
{
 long long fsize = 0;
 QString str;
 
 for (int i=0; i<numImages; i++)
  if (allIconsInfo[i].filterOK || (!usingFilter))
   fsize += allIconsInfo[i].fileSize;

 if (fsize<1000)
  str = QString("%1").arg(fsize);
 else if (fsize<1000000)
  str.sprintf("%lld.%03lld ", fsize/1000, fsize%1000);
 else if (fsize<1000000000)
  str.sprintf("%lld.%03lld.%03lld", fsize/1000000, (fsize/1000)%1000, fsize%1000);
 else
  str.sprintf("%lld.%03lld.%03lld.%03lld", fsize/1000000000, (fsize/1000000)%1000, (fsize/1000)%1000, fsize%1000);

 str.append(" "+ApaintConfig::getText(BYTES_MSG));
 return(str);
}

/******************QPtrListIcon*******************/
int QPtrListIcon::compareItems( QPtrCollection::Item s1, QPtrCollection::Item s2 )
{
 imageInfo *e1, *e2;
 QString pfx1, pfx2;
 int i1, i2;
 i1 = i2 = 0;
 e1 = (imageInfo*)s1;
 e2 = (imageInfo*)s2;

 switch(selectedSort)
  {
    case NameSort:
      return(strcmp(e1->Fname, e2->Fname));
    break;
    case FileTypeSort:
      pfx1 = getPfx(e1->Fname);
      pfx2 = getPfx(e2->Fname);
      return(strcmp(pfx1.ascii(), pfx2.ascii()));
    break;
    case SizeSort:
       i1=e1->imageX;
       i2=e2->imageX;
       if (i1==i2)
       {
         i1=e1->imageY;
         i2=e2->imageY;
       }
    break;
    case SizeXSort:
      i1 = e1->imageX;
      i2 = e2->imageX;
    break;
    case SizeYSort:
      i1 = e1->imageY;
      i2 = e2->imageY;
    break;
    case FileSizeSort:
      i1 = e1->fileSize;
      i2 = e2->fileSize;
    break;
    case DTimeSort:
      i1 = e1->modifiedTime;
      i2 = e2->modifiedTime;
    break;
    case ColorNumSort:
      i1 = e2->usedColors;
      if (i1 == 0) i1 = 300;    // if true color put it in the back
      i2 = e1->usedColors;
      if (i2 == 0) i2 = 300;      
    break;
  }
 if (i1 == i2)
  return(strcmp(e1->Fname, e2->Fname));
 else if (i1>i2)
  return -1;
 else
  return 1;
}
