/***************************************************************************
                          freehandselect.cpp  -  description
                             -------------------
    begin                : Mon Aug 18 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qpainter.h> 
#include "toolFreehandSelect.h"
#include "apaintUtil.h"
#include "apaint.h"

extern char *getCommonPNG(const char *pixName);

FreeHandSelect::FreeHandSelect(ToolType tooltype, ApaintDoc *docp, PropObject *propObj):ToolBase(tooltype, propObj)
{
 cur=QCursor(ArrowCursor);
 setToggleBtn(QPixmap(getCommonPNG("select-lasso")), tr("Select"));
 leftBtnDown = rightBtnDown = drawStarted = false;
 state = DrawState;
 mpos = OutArea;
 doc=docp;
 startx = starty = lastx = lasty = firstx = firsty = 0;
 if (propObj)
  setUsepWidget(true);
 ptsList.clear();
 ptsList.setAutoDelete(true);
 allCur = QCursor(Qt::SizeAllCursor);
 mainCur = QCursor();
 connect(doc, SIGNAL(layerChanged(Layer *)), this, SLOT(layerChangedSlot(Layer *)));
}

FreeHandSelect::~FreeHandSelect()
{
}


void FreeHandSelect::layerChangedSlot(Layer *layer)
{
 beginTool();
}

void FreeHandSelect::begin()
{
 ToolBase::begin();
 beginTool();
}

void FreeHandSelect::end()
{
 ToolBase::end();
 endTool();
 emit unsetCursor();
}

void FreeHandSelect::beginTool()
{
 if (doc->getLayerType()==SelectionLayer || doc->getLayerType()==ImageLayer)
  {
   state = MoveState;
   layerCreated = true;
   setAutoScroll(true);
   autoScrollNoButton=FALSE;
  }
 else
  {   
   state = DrawState;
   layerCreated=false;
   drawStarted = false;   
  } 
}

void FreeHandSelect::endTool()
{
 QPoint *pt;
 int i;

 if (state == DrawState)
  {
   if (!ptsList.count()) return;
   borderPts = QPointArray(ptsList.count());
   for (pt = ptsList.first(), i=0; pt; pt = ptsList.next(), i++)
     borderPts.setPoint(i, pt->x(), pt->y());
   ptsList.clear();
   QRect r = borderPts.boundingRect().normalize();
   borderPts = QPointArray();
   doc->refreshLayer(r);
  }
 doc->setLayerResizeable(false);
}

void FreeHandSelect::createLayer()
{
 QPoint *pt;
 int i;

 borderPts = QPointArray(ptsList.count());
 for (pt = ptsList.first(), i=0; pt; pt = ptsList.next(), i++)
   borderPts.setPoint(i, pt->x(), pt->y());
 ptsList.clear();
 
 doc->createNewLayer(borderPts, SelectionLayer, Freehand);
 if (doc->removeNullLayer()) return;
 /*for CreateLayer undo data is always taken from mainLayer so we can call it after creation*/
 doc->putUndoStack(CreateLayer);
 drawStarted = false;
 state = MoveState;
 layerCreated = true;
}

void FreeHandSelect::drawScaleLine(QPainter *p, int x1, int y1, int x2, int y2)
{
  p->setRasterOp(Qt::XorROP);
  p->setPen(QPen(QColor(128, 128,0),QMAX(2, (int)(zoomX/2)),Qt::DashLine));
  p->drawLine((int) (x1*zoomX)+(int)(zoomX/2), (int) (y1 *zoomY)+(int)(zoomY/2), (int) (x2*zoomX)+(int)(zoomX/2), (int) (y2 *zoomY)+(int)(zoomY/2));
}


void FreeHandSelect::refreshPoly(QPainter *p)
{
 if (state == DrawState)
  {
   int x1, y1, x2, y2, i, count;
   QPoint *pt;

   p->setRasterOp(Qt::XorROP);
   p->setPen(QPen(QColor(128, 128,0),QMAX(2, (int)(zoomX/2)),Qt::DashLine));

   pt = ptsList.first();
   count = ptsList.count();
   if (!count) return;

   x1 = (int)(pt->x()*zoomX);
   y1 = (int)(pt->y()*zoomY);
   for (i =1; i<count; i++)
   {
     pt = ptsList.next();
     x2 = (int)(pt->x()*zoomX);
     y2 = (int)(pt->y()*zoomY);
     p->drawLine(x1, y1, x2, y2);
     x1 = x2;
     y1 = y2;
   }
  } 
}

void FreeHandSelect::beforeRefresh(QPainter *p)
{
  if (state == DrawState && drawStarted)
  {
   refreshPoly(p);
   drawScaleLine(p, startx, starty, lastx, lasty);
  } 
}

void FreeHandSelect::afterRefresh(QPainter *p)
{
  if (state == DrawState && drawStarted)
  {
   refreshPoly(p);
   drawScaleLine(p, startx, starty, lastx, lasty);
  }
}

void FreeHandSelect::doOp(QPainter *pc, int x, int y)
{
 if (state == DrawState)
 {
  QPoint *pt  = new QPoint(x, y);
  ptsList.append(pt);
  drawScaleLine(pc, lastx, lasty, x, y);
 }
 else if (state == MoveState && mpos == InArea)
 {
  int dx, dy;

  dx=lastx-x;
  dy=lasty-y;
  ApaintDocUtil::moveCurLayer(doc, dx, dy, QMAX(2, (int)(zoomX/2)));
  doc->setLayerResizeable(false);   
 }
 
 lastx = startx = x;
 lasty = starty = y;
}

void FreeHandSelect::selectOp(QPainter *pc, int x, int y)
{
 QCursor curL;
   
 if (state == DrawState && drawStarted)
  {
   drawScaleLine(pc, startx, starty, lastx, lasty);
   drawScaleLine(pc, startx, starty, x, y);
   lastx = x;
   lasty = y;
  }
 else if (state == MoveState && layerCreated)
 {
  if (doc->isLayerPointVisible(x, y))
   {
    mpos = InArea;
    curL = allCur;
   }
  else
   {
    mpos = OutArea;
    curL = cur;
   }
  if (mainCur.handle()!=curL.handle())
   {
    mainCur = curL;
    emit setCursor(mainCur);
   }
 } 
}

bool FreeHandSelect::pointInImage(int x, int y)
{
 QRect pos = doc->getMainLayer()->layerPos();
 if (pos.contains(x,y))
  return true;
 else
  return false;
}

void FreeHandSelect::mousePressed(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
 int x=(int)(e->x()/zoomX);
 int y=(int)(e->y()/zoomY); 

 pc->setRasterOp(Qt::NotROP);
 pc->setPen(QPen(Qt::DashLine));

 if (e->button() == LeftButton && pointInImage(x,y))
  {
   leftBtnDown = true;
   apaint->selectionMoving(true, doc->getLayer());
  }
 else if (e->button() == RightButton)
  rightBtnDown = true;

 if (state == DrawState)
  {
    if (leftBtnDown)
    {
      if (!drawStarted)
       {
        firstx = x;
        firsty = y;
       }
      drawStarted = true;
      startx = lastx = x;
      starty = lasty = y;
      doOp(pc, x, y) ;
      autoScrollNoButton=true;
    }
    else if (rightBtnDown && drawStarted)
    {
     if (ptsList.count()<3)
      {
       ptsList.clear();
       drawStarted = false;
       drawScaleLine(pc, startx, starty, x, y);
       drawScaleLine(pc, firstx, firsty, startx, starty);
       autoScrollNoButton=false;
       return;
      }
     drawScaleLine(pc, startx, starty, lastx, lasty);
     lastx = startx;
     lasty = starty;
     doOp(pc, firstx, firsty);
     createLayer();
    }
  }
 else if (state == MoveState)
  {
   lastx = x;
   lasty = y;

   if (mpos==InArea)
    doc->putUndoStack(ChangeLayerPos);
   else
    if (mpos == OutArea && rightBtnDown)
     ApaintDocUtil::destroyCurLayer(doc);
   autoScrollNoButton=false;
  }
}

void FreeHandSelect::mouseMoved(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
 int x=(int)(e->x()/zoomX);
 int y=(int)(e->y()/zoomY);

 pc->setRasterOp(Qt::XorROP);
 pc->setPen(QPen(Qt::SolidLine));
 /*selection must be smaler than main image*/
 QRect pos = doc->mainLayer->layerPos();
 if (x<0) x=0;
 if (y<0) y=0;
 if (x>pos.width()) x = pos.width()-1;
 if (y>pos.height()) y = pos.height()-1;
 
 if (leftBtnDown)
  {
   if (state == DrawState)
    drawPointer(pc, e->x(), e->y());
   doOp(pc, x, y);
   if (state != DrawState)
    drawPointer(pc, e->x(), e->y());
  }
 else
  {
   drawPointer(pc, e->x(), e->y());
   selectOp(pc, x, y);
  }
}

void FreeHandSelect::mouseReleased(QMouseEvent *e, QPainter *pc, QPainter *pi)
{   
 leftBtnDown = rightBtnDown = false;
 apaint->selectionMoving(false, doc->getLayer());
}

void FreeHandSelect::keyPress(QKeyEvent *e, QPainter *pc, QPainter *pi)
{
  int dx, dy, inc;
  bool keyOK = false;

  dx = 0; dy =0;
  if (e->state() == ShiftButton)
   inc = 10;
  else
   inc = 1;
  if (e->key() == Qt::Key_Left) {dx+=inc; keyOK = true;}
  if (e->key() == Qt::Key_Right) {dx-=inc; keyOK = true;}
  if (e->key() == Qt::Key_Up) {dy+=inc; keyOK = true;}
  if (e->key() == Qt::Key_Down) {dy-=inc; keyOK = true;}

  if (!keyOK) return;

  if (state==MoveState)
   {
    ApaintDocUtil::moveCurLayer(doc, dx, dy, QMAX(2, (int)(zoomX/2)));
    lastx+=dx;
    lasty+=dy;
   }
}

