/***************************************************************************
                          toolsctrl.cpp  -  description
                             -------------------
    begin                : Sun Feb 2 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <typeinfo>
#include "toolMagicWand.h"
#include "toolsctrl.h"
#include "apaintdrawrepeat.h"
#include "apaintdoc.h"

#define SCROLL_BY    20

static const char *colorPicker[]={
"16 16 10 1",
". c None",
"# c #000000",
"a c #004000",
"b c #303030",
"f c #585858",
"h c #808080",
"d c #a0a0a0",
"c c #c3c3c3",
"g c #dcdcdc",
"e c #ffffff",
"............#ab.",
"...........#cd#b",
"..........#cd###",
"........##cd##a#",
".......#ccd##a#.",
"........#dd#a#..",
".......#ef###...",
"......#egh###...",
".....#egh#.#....",
"....#egh#.......",
"...#egh#........",
"..#egh#.........",
".#egh#..........",
".#gh#...........",
"#g##............",
"##.............."};

ToolsCtrl::ToolsCtrl(DrawInRepeat *drawRepIn, ApaintDoc *docp)
{  
   selectedTool=0;
   toolIsActive=dragStart=false;
   shiftPressed=false;
   colSel=QCursor(QPixmap(colorPicker), 0, 15);
   zoomX=zoomY=1;
   aspectX = aspectY = 1;
   pointerPosX = pointerPosY = -1;
   sendEvents = true;
   toolIsBeingChanged = false;
   globX = globY = 0;
   gxpos = gypos = 0;
   usingMask = false;
   changePointer = true;
   Q_CHECK_PTR(drawRepIn);
   drawRep = drawRepIn;
   Q_CHECK_PTR(docp);
   doc = docp;
}

ToolsCtrl::~ToolsCtrl()
{
}

void ToolsCtrl::enableToolProp(PropObject *pobj, bool enable)
{
  if (selectedTool)
  {
    if (pobj == selectedTool->pObject())
     selectedTool->enablePropWindow(enable);
  }
}

void ToolsCtrl::setImgPainter(QPainter *p)
{
 selLayer = doc->getLayer();
 if (!selLayer) return;
 p->setPen(mainPen);
 p->setBrush(mainBrush);
 p->translate(-selLayer->drawAreaPos().x(), -selLayer->drawAreaPos().y());

}

void ToolsCtrl::brushChgSlot(const QBrush &b)
{
  mainBrush=b;
}

void ToolsCtrl::lineSizeChgSlot(const QPen &p)
{
  mainPen = p;
}

ToolBase *ToolsCtrl::getActiveTool()
{
  return selectedTool;
}

void ToolsCtrl::registerTool(ToolBase *tool)
{      
   toolIsBeingChanged = true;
   selLayer = doc->getLayer();
   if (!selLayer) 
   {
     tool->end();
     toolIsActive=false;
     toolIsBeingChanged = false;
     return;
   }
   if (!tool)
   {
     selectedTool=0;
     toolIsBeingChanged = false;
     return;
   }
   if (tool->getType() == ActionTool)
   {
      toolIsBeingChanged = false;
      if (sendEvents)
        tool->doAction();
      return;
   }
   if (selectedTool)
   {
     unregisterTool();
   }   
   selectedTool=tool;
   emit setCur(selectedTool->getCursor());
   selectedTool->setZoom(zoomX, zoomY);
   selectedTool->setAspectRatio(aspectX, aspectY);
   selectedTool->showCommonProp(true);
   emit setAutoScroll(selectedTool->usingAutoScroll());
   toolIsActive=true;
   toolIsBeingChanged = false;
   selectedTool->begin();
   if (selectedTool->usePropertyWin())
     selectedTool->showPropWindow(true);
}

void ToolsCtrl::unregisterTool()
{
  if (!toolIsActive) return;
  emit unsetCur();  
  selectedTool->end();
  selectedTool->showCommonProp(false);
  if (selectedTool->usePropertyWin())
    selectedTool->showPropWindow(false);
  selectedTool=0;
  toolIsActive=false;
  emit setAutoScroll(false);
}

void ToolsCtrl::slotAspectChanged(double ax, double ay)
{
 aspectX = ax;
 aspectY = ay;
 if (selectedTool)
  selectedTool->setAspectRatio(aspectX, aspectY);
}

void ToolsCtrl::slotZoomChanged(double zx, double zy)
{ 
  zoomX=zx;
  zoomY=zy;
  if (toolIsActive)
   selectedTool->setZoom(zoomX, zoomY);
}

void ToolsCtrl::getColor(int xs, int ys)
{
  uchar *src;
  uchar col8;
  int x, y;

  selLayer = doc->getLayer();
  if (!selLayer) return;
  
  QRect pos = selLayer->layerPos();
  drawRep->transPoint((int)(xs/zoomX), (int)(ys/zoomY), &x, &y);
  if (!pos.contains(x,y))
   return;
  x-=pos.x();
  y-=pos.y();
  if (selLayer->getImageType() == LOAD_TRUE_COL)
  {
   src=(uchar *)selLayer->imageBuf.scanLine(y)+x*3;   
   emit curRGB(*src, *(src+1), *(src+2), 0);
  }
  else if (selLayer->getImageType() == LOAD_CMAP)
  {
   QRgb *ctab = selLayer->imageBuf.colorTable(); 
   col8=*(selLayer->imageBuf.scanLine(y)+x);
   emit curRGB(qRed(ctab[col8]), qGreen(ctab[col8]), qBlue(ctab[col8]), col8+1);
  }
}


void ToolsCtrl::setColor(int xs, int ys)
{
  uchar *src;
  uchar col8;
  int x, y;

  selLayer = doc->getLayer();
  if (!selLayer) return;

  QRect pos = selLayer->layerPos();
  drawRep->transPoint((int)(xs/zoomX), (int)(ys/zoomY), &x, &y);  
  if (!pos.contains(x,y)) return;
  x-=pos.x();
  y-=pos.y();
  if (selLayer->getImageType() == LOAD_TRUE_COL)
  {
   src=(uchar *)selLayer->imageBuf.scanLine(y)+x*3;  
   emit setColorSig(qRgb(*src, *(src+1), *(src+2)));
  }
  else if (selLayer->getImageType() == LOAD_CMAP)
  {
   col8=*(selLayer->imageBuf.scanLine(y)+x);
   emit setColorSigIndx(col8);
  }
}

void ToolsCtrl::mousePressSlot(QMouseEvent *e, QPainter *pc)
{
 QCursor q;

  selLayer = doc->getLayer();
  if (!selLayer) return;

  if (e->button()==LeftButton && e->state() == ShiftButton)
   setColor(e->x(), e->y());
  else
   if (toolIsActive)
    {
     if (selectedTool->getType()==DrawTool)
      {
       emit drawStarted();
       QPainter pi(selLayer->getPixmap());
       setImgPainter(&pi);
       if (selLayer->layerPos().contains((int)(e->x()/zoomX), (int)(e->y()/zoomY)) && sendEvents)
	selectedTool->mousePressed(e, pc, &pi);
      }
     else
      if ( (selectedTool->getType()==SelectionTool) && sendEvents)
       selectedTool->mousePressed(e, pc, pc);

     /* handles moving of area with middle mouse button */
     if (e->button()==MidButton)
       {
	scrollStartX=e->x();
	scrollStartY=e->y();
	dragStart=true;
	q=QCursor(SizeAllCursor);
	emit setCur(q);
       }
   }
}
                            
void ToolsCtrl::mouseMoveSlot(QMouseEvent *e, QPainter *pc)
{
 int deltaX,deltaY, xpos, ypos;

 selLayer = doc->getLayer();
 if (!selLayer) return;

 globX = e->x();
 globY = e->y();
 xpos  = (int)((double)e->x()/zoomX);
 ypos =  (int)((double)e->y()/zoomY);
 if (gxpos == xpos && gypos == ypos) return;
 gxpos = xpos;
 gypos = ypos;
 emit coordsPos(xpos, ypos);
 if (e->state() == ShiftButton)
  {
   getColor(e->x(), e->y());
   return;
  }
 if (toolIsActive)
  {
   if (sendEvents)
   {
   if (pointerPosX == -1 && pointerPosY == -1)
	{
         pointerPosX = globX;
         pointerPosY = globY;
	}
   else
	selectedTool->drawPointer(pc, pointerPosX, pointerPosY);
   pointerPosX = globX;
   pointerPosY = globY;
   }

   if (selectedTool->getType()==DrawTool)
    {
     QPainter pi(selLayer->getPixmap());
     setImgPainter(&pi);
     if (sendEvents)
      {
       changePointer = false;
       selectedTool->mouseMoved(e, pc, &pi);
       selectedTool->drawPointer(pc, pointerPosX, pointerPosY);
       changePointer=true;
      }
    }
   else
    if ((selectedTool->getType()==SelectionTool))
    {
       changePointer = false;
     if (sendEvents)
     {
    	 selectedTool->mouseMoved(e, pc, pc);
     }

	 if (typeid(*selectedTool) == typeid(MagicWandTool))
	  selectedTool->drawPointer(pc, pointerPosX, pointerPosY);

     changePointer = true;
    }
   if (dragStart)
    {
     deltaX=scrollStartX-e->x();
     deltaY=scrollStartY-e->y();
     emit scrollBySig(deltaX,deltaY);
    }
  }
}

void ToolsCtrl::mouseReleaseSlot(QMouseEvent *e, QPainter *pc)
{
 selLayer = doc->getLayer();
 if (!selLayer) return;

 if (toolIsActive)
  {
   if (selectedTool->getType()==DrawTool)
    {
     QPainter pi(selLayer->getPixmap());
     setImgPainter(&pi);
     if (sendEvents)
      selectedTool->mouseReleased(e, pc, &pi);
    }
   else
    if (selectedTool->getType()==SelectionTool)
     {
      if (sendEvents)
       selectedTool->mouseReleased(e, pc, pc);
     }
   if (dragStart)
    {
     dragStart=false;
     emit unsetCur();
    }
  }
 else
  if (dragStart) /* we leave it here just in case we got stuck in moving ... */
   {
    dragStart=false;
    emit unsetCur();
   }
}

void ToolsCtrl::repaintSlot(QPainter *pc)
{
 selLayer = doc->getLayer();
 if (!selLayer) return;

 if (toolIsActive)
  {
   if (selectedTool->getType()==DrawTool)
	{
	 QPainter pi(selLayer->getPixmap());
	 setImgPainter(&pi);
	 if (sendEvents)
	  selectedTool->repaint(pc, &pi);
	}
  }
}

void ToolsCtrl::mouseDoubleClickSlot(QMouseEvent *e, QPainter *pc)
{
  selLayer = doc->getLayer();
  if (!selLayer) return;

  if (toolIsActive)
  {
   if (selectedTool->getType()==DrawTool)
    {
     QPainter pi(selLayer->getPixmap());
     setImgPainter(&pi);
     if (selLayer->layerPos().contains((int)(e->x()/zoomX), (int)(e->y()/zoomY)) && sendEvents)
        selectedTool->mouseDoubleClick(e, pc, &pi);
    }
   else if (selectedTool->getType()==SelectionTool)
    {
     if (sendEvents) 
      selectedTool->mouseDoubleClick(e, pc, pc);
    } 
  }
}

void ToolsCtrl::keyPressSlot(QKeyEvent *e, QPainter *pc)
{
  selLayer = doc->getLayer();
  if (!selLayer) return;

  if (e->key()==Qt::Key_Shift && e->state() != ControlButton && e->state() != AltButton)
  {
   shiftPressed=true;
   emit setCur(colSel);
  }
  else if (e->key() == Key_Control)
  {
    if (shiftPressed)
    {
      shiftPressed=false;
      emit unsetCur();
      emit clearSize();   
    }
  }
  else if (e->key() == Key_Alt || e->key() == Key_Meta)
  {
    if (shiftPressed)
    {
      shiftPressed=false;
      emit unsetCur();
      emit clearSize();
    }
  }
  else if (e->key()==Qt::Key_Space)
  {
    setColor(globX, globY);
  }    
  if (toolIsActive)
  {
    if (selectedTool->getType()==DrawTool)
    {
     QPainter pi(selLayer->getPixmap());
     setImgPainter(&pi);
     
     switch(e->key())
     {
       case Qt::Key_Left:
            emit scrollBySig(-SCROLL_BY, 0);
              break;
        case Qt::Key_Right:
             emit scrollBySig(SCROLL_BY, 0);
                break;
        case Qt::Key_Up:
             emit scrollBySig(0, -SCROLL_BY);
                break;
        case Qt::Key_Down:
             emit scrollBySig(0, SCROLL_BY);
                break;
        default:
            if (sendEvents)
             selectedTool->keyPress(e, pc, &pi);
      }
    }
    else if (selectedTool->getType()==SelectionTool)
    {
      if (sendEvents)
        selectedTool->keyPress(e, pc, pc);
    }
  }  
}

void ToolsCtrl::keyReleaseSlot(QKeyEvent *e, QPainter *pc)
{
  selLayer = doc->getLayer();
  if (!selLayer) return;

  if (e->key()==Qt::Key_Shift)
  {
   shiftPressed=false;
   emit unsetCur();
   emit clearSize();   
  }
  if (toolIsActive)
  {
    if (selectedTool->getType()==DrawTool)
    {
     QPainter pi(selLayer->getPixmap());
     setImgPainter(&pi);
     if (sendEvents)
      selectedTool->keyRelease(e, pc, &pi);
    }
    else if (selectedTool->getType()==SelectionTool)
    {
      if (sendEvents)
        selectedTool->keyRelease(e, pc, pc);
    }
  }  
}

void ToolsCtrl::leaveEventSlot(QEvent* e, QPainter *pc)
{
 refreshPointer(pc);
 changePointer = false;
}

void ToolsCtrl::enterEventSlot(QEvent* e, QPainter *pc)
{
  changePointer = true;
  refreshPointer(pc);
}

void ToolsCtrl::beforeRefreshSlot(QPainter* pc)
{
 refreshPointer(pc);
 if (toolIsActive)
  selectedTool->beforeRefresh(pc);
}

void ToolsCtrl::afterRefreshSlot(QPainter* pc)
{
 refreshPointer(pc);
 if (toolIsActive)
  selectedTool->afterRefresh(pc);
}

void ToolsCtrl::refreshPointer(QPainter *p)
{
 if (toolIsActive && changePointer)
 {   
   selectedTool->drawPointer(p, pointerPosX, pointerPosY);
 }
  
}

void ToolsCtrl::setSendEvents(bool on)
{
 sendEvents = on; 
}

void ToolsCtrl::getGlobalMousePos(int &xpos, int &ypos)
{
  xpos = globX;
  ypos = globY;
}


