
#include "widgetGuide.h"
#include <qpainter.h>
#include <qpixmap.h>
#include <stdlib.h>

Guide::Guide(ApaintView* parent, ApaintDoc* _doc, Orientation _guideOrientation)
: QWidget(parent->viewport(), "")
{
  this->guideOrientation = _guideOrientation;
  
  view = (ApaintView*) parent;
  doc = _doc;
  
  if (guideOrientation == Vertical)
  {
    setMaximumWidth(1);
    this->resize(1, view->viewport()->height());;
    setCursor(Qt::sizeHorCursor);
  }
  else 
  {
    setMaximumHeight(1);
    this->resize(view->viewport()->width(),1);
    setCursor(Qt::sizeVerCursor);
  }
  offsetX = view->getOffsetX();
  offsetY = view->getOffsetY();
  
  zoomX = view->getScaleX();
  zoomY = view->getScaleY();
  
  posX = posY = 0;
  
  this->setColors();
  
  
  setBackgroundColor(bcgColor);

  
}

Guide::~Guide()
{

}

void Guide::mousePressEvent(QMouseEvent* event)
{
    QPoint point = event->globalPos();
    lastX = point.x();
    lastY = point.y();
}

void Guide::mouseReleaseEvent(QMouseEvent* event)
{
    QWidget::mouseReleaseEvent(event);
}

void Guide::mouseDoubleClickEvent(QMouseEvent* event)
{
    delete this;
}

void Guide::mouseMoveEvent(QMouseEvent *e)
{
    QPoint point = e->globalPos();

    int x = point.x();
    int y = point.y();
    
    
    int dx = x - lastX;
    int dy = y - lastY;

    QPoint guidePos = this->pos();
    
    lastX = x;
    lastY = y;
    
    if (guideOrientation == Qt::Horizontal)
    {
        if (dy < 0)
        {
            this->move(guidePos.x(), QMAX(guidePos.y()+dy, view->viewport()->rect().y()));
        }
        else
        {
            this->move(guidePos.x(), QMIN(guidePos.y()+dy, view->viewport()->rect().bottom()));

        }
    }
    else
    {
        if (dx < 0)
        {
            this->move(QMAX(guidePos.x()+dx, view->viewport()->rect().x()), guidePos.y());
        }
        else
        {
            this->move(QMIN(guidePos.x()+dx, view->viewport()->rect().right()), guidePos.y());
        }
    }
    update();
    setBasePos();
    emit movedSig();
}


void Guide::paintEvent(QPaintEvent* event)
{
    QRect rect = event->rect();
    
    if (guideOrientation == Vertical)
    {
      setMaximumWidth(1);
      this->resize(1, view->height());
      setCursor(Qt::SizeHorCursor);
    }
    else 
    {
      setMaximumHeight(1);
      this->resize(view->width(),1);
      setCursor(Qt::sizeVerCursor);
    }
    
    QPainter dbPainter(this);
    QPen pen(lineColor, 2, Qt::DashDotDotLine);
    
    dbPainter.setPen(pen);
    if (guideOrientation == Horizontal)
    {
      dbPainter.drawLine(0, 0, rect.right(), 0);
      move(view->viewport()->rect().x(), pos().y());
    }
    else
    {
      dbPainter.drawLine(0, rect.y(), 0, rect.bottom());
      move(pos().x(), view->viewport()->rect().y());
    }
}

void Guide::slotZoomChanged(double zx, double zy)
{ 
  zoomX=zx;
  zoomY=zy;
  
  if(guideOrientation == Qt::Horizontal)
      move(pos().x(), (int)floor((double)(posY)*zoomY) - offsetY );
  else
      move((int)floor((double)(posX)*zoomX) - offsetX, pos().y() );
}

void Guide::viewMoving(int x, int y)
{
  
  offsetX = x;
  offsetY = y;
  
  update();
  
  //without this call view is not updated correctly
  view->viewport()->update();
}

void Guide::setBasePos()
{
    //calculate position on non-zoomed image
    posX = (int)floor((double)(offsetX + pos().x())/zoomX);
    posY = (int)floor((double)(offsetY + pos().y())/zoomY);
}

void Guide::repaintSlot()
{
  update();
}

void Guide::setColors()
{
   int num = doc->layerNumColors();
  
   int newRGB[3] = {0, 0, 0};
   int bestRGB[3] = {0, 0, 0};
   int bcgRGB[3] = {0, 0, 0};
   int i, j, diff, minDiff, bestDiff;
   QRgb* ctab; 

   //putUndoStack(ChangeColorTable);
   ctab = doc->getLayer()->imageBuf.colorTable();

   /* create 750 random colors and pick the one which is most different from all the colors in existing color palette */
   bestDiff = 0;
   for (i=0; i<750; i++)
    {
     newRGB[0] = rand()&255;
     newRGB[1] = rand()&255;
     newRGB[2] = rand()&255;
     minDiff=INT_MAX;
     for (j=0; j<num; j++)
      {  /* we make green the most important for color difference, as it contributes most to the lightness */
       diff  = (qRed(ctab[j])-newRGB[0])*(qRed(ctab[j])-newRGB[0])*3;
       diff += (qGreen(ctab[j])-newRGB[1])*(qGreen(ctab[j])-newRGB[1])*4;
       diff += (qBlue(ctab[j])-newRGB[2])*(qBlue(ctab[j])-newRGB[2])*2;
       if (diff<minDiff)
	minDiff = diff;
      }
     if (minDiff>bestDiff)
      {
       bestDiff=minDiff;
       bestRGB[0]=newRGB[0];
       bestRGB[1]=newRGB[1];
       bestRGB[2]=newRGB[2];
      }
     }

       bcgRGB[0]=bestRGB[0];
       bcgRGB[1]=bestRGB[1];
       bcgRGB[2]=bestRGB[2];
    
     
     bcgColor = QColor(bestRGB[0], bestRGB[1], bestRGB[2]);
   //now do the same thing for line color but with the background color included into pallete
    bestDiff = 0;
    for (i=0; i<750; i++)
    {
      newRGB[0] = rand()&255;
      newRGB[1] = rand()&255;
      newRGB[2] = rand()&255;
      minDiff=INT_MAX;
      
      
      for (j=0; j<num; j++)
      {  /* we make green the most important for color difference, as it contributes most to the lightness */
	diff  = (qRed(ctab[j])-newRGB[0])*(qRed(ctab[j])-newRGB[0])*3;
	diff += (qGreen(ctab[j])-newRGB[1])*(qGreen(ctab[j])-newRGB[1])*4;
	diff += (qBlue(ctab[j])-newRGB[2])*(qBlue(ctab[j])-newRGB[2])*2;
	if (diff<minDiff)
	  minDiff = diff;
      }
      
	diff  = (bcgRGB[0]-newRGB[0])*(bcgRGB[0]-newRGB[0])*3;
	diff += (bcgRGB[1]-newRGB[1])*(bcgRGB[1]-newRGB[1])*4;
	diff += (bcgRGB[2]-newRGB[2])*(bcgRGB[2]-newRGB[2])*2;
	if (diff<minDiff)
	  minDiff = diff;

      if (minDiff>bestDiff)
      {
	bestDiff=minDiff;
	bestRGB[0]=newRGB[0];
	bestRGB[1]=newRGB[1];
	bestRGB[2]=newRGB[2];
      }

    lineColor = QColor(bestRGB[0], bestRGB[1], bestRGB[2]);
    
  }
}

void Guide::newDocSlot()
{
  delete this;
}

