/***************************************************************************
                          widgetIconView.cpp  -  description
                             -------------------
    begin                : Wed Apr 21 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdlib.h>
#include <qpainter.h>
#include <qtooltip.h>
#include "apaintio.h"
#include "apaintconfig.h"
#include "widgetIconView.h"
#include "iconViewData.h"

static int loadingNumFound=0;

static void displayIconReadProgress(int numFound)
{
 loadingNumFound = numFound;
/*
 QString label;

 label = ApaintConfig::getText(FOUND_MSG) + QString(" %1").arg(numFound);

 printf("%s\n", label.ascii());

*/

/*
   {
    QRect clean = QRect(0, 0, 200, 200);
    p->fillRect(clean, Qt::white);
    p->drawText(50, 50, label);
   }
*/
}


 
IconListView::IconListView(QWidget *parent, const char* name):QScrollView(parent, name, WRepaintNoErase|WResizeNoErase|WStaticContents)
{
 iconx = icony = 120 + 2;
 iconBelowMouse = 0;
 viewport()->setBackgroundMode(NoBackground);
 viewport()->setMouseTracking(true);
 iconsLoaded = false;
 setFocusPolicy(QWidget::StrongFocus);
 viewport()->setFocusPolicy(QWidget::StrongFocus);
 registerBrowseIconProgress(displayIconReadProgress);
}

IconListView::~IconListView()
{

}

void IconListView::setIconList(IconList *ilistp)
{
 if (!ilistp) return;
 iList = ilistp;
 iconsLoaded = true;
 setupContents(false);
}

void IconListView::setupContents(bool refreshAll)
{
  if (!iconsLoaded || !iList) return;
  int ncol = QMAX(1, width() / iconx);
  int oldNCol = iList->getNumCol();

  iList->setSize(ncol, iconx, icony);
  vwidth = iList->width();
  vheight = iList->height();
  resizeContents(vwidth, vheight);
  if (ncol!=oldNCol || refreshAll)
   repaintContents();
}

void IconListView::refresh()
{
  repaintContents(); 
}

void IconListView::changeSelected(int newSel, bool visible)
{
 IconData *icData;
 QRect oldPos;

 icData = iList->getIconData(iList->selectedIcon());
 if (icData)
  oldPos = icData->getIconPos() | icData->getTextPos();
 iList->setSelectedIcon(newSel);
 icData = iList->getIconData(iList->selectedIcon());
 if (icData)
 {
  QRect newPos = icData->getIconPos() | icData->getTextPos();
  repaintContents(newPos);
  if (visible)
   ensureVisible(newPos.left(), newPos.bottom());
 }
 repaintContents(oldPos);
 emit iconSelectionChanged();
}

void IconListView::goLeft()
{
 changeSelected(iList->selectedIcon()-1, true);
}

void IconListView::goRight()
{
 changeSelected(iList->selectedIcon()+1, true);
}

void IconListView::goUp()
{
 changeSelected(iList->selectedIcon()-iList->getNumCol(), true);
}

void IconListView::goDown()
{
 changeSelected(iList->selectedIcon()+iList->getNumCol(), true);
}

void IconListView::pageUp()
{
 int numPage;
 IconData *icData = iList->getIconData(0);
 if (icData)
 {
  QRect pos = icData->getIconPos() | icData->getTextPos();
  numPage = QMAX(1, visibleHeight()/pos.height());
  changeSelected(iList->selectedIcon()-iList->getNumCol()*numPage, true);
 }
}

void IconListView::pageDown()
{
 int numPage;
 IconData *icData = iList->getIconData(0);
 if (icData)
 {
  QRect pos = icData->getIconPos() | icData->getTextPos();
  numPage = QMAX(1, visibleHeight()/pos.height());
  changeSelected(iList->selectedIcon()+iList->getNumCol()*numPage, true);
 }
}

void IconListView::deleteImage()
{
  emit deleteImageSig();
}

void IconListView::renameImage()
{
  emit renameImageSig();
}

void IconListView::viewportResizeEvent(QResizeEvent *re)
{
  QScrollView::viewportResizeEvent(re);
  setupContents(false);
}

void IconListView::drawContents(QPainter *p, int clipx, int clipy, int clipw, int cliph)
{
  if (!clipw || !cliph) return;
  QToolTip::hide();
  p->setBackgroundColor(QColor(white));
  if (clipx>vwidth || clipy>vheight || !iconsLoaded)
   {
    p->fillRect(clipx,clipy,clipw,cliph,white);
    return;
  }

  QFontMetrics fm = fontMetrics();
  ApaintIO *io = ApaintIO::getObject();
  if (io)
   io->fastIconDraw(iList, p, fm, clipx, clipy, clipw, cliph);

  if (clipx+clipw>vwidth)
   {
    int overw = clipx+clipw-vwidth;
    p->fillRect(vwidth, clipy, overw, cliph, white);
   }
  if (clipy+cliph>vheight)
   {
    int overh = clipy+cliph-vheight;
    p->fillRect(clipx, vheight, clipw, overh, white);
   }
}

void IconListView::contentsMousePressEvent(QMouseEvent *e )
{
   if (!iconsLoaded) return;

   int realx = e->pos().x();
   int realy = e->pos().y();

   int x =  realx / iconx;
   int y =  realy / (iList->getTextHeight()+icony);

   if (!iList->validPos(x, y)) return;

  changeSelected(y*iList->getNumCol()+x, false);
  QRect renRect = iList->getIconData(iList->selectedIcon())->getTextPos();
  // left half should be active
  renRect = QRect(renRect.x(), renRect.y(), renRect.width()/2, renRect.height());
  if (renRect.contains(realx, realy))
   renameImage();
}

void IconListView::contentsMouseDoubleClickEvent (QMouseEvent *e )
{
   if (!iconsLoaded) return;
   emit iconSelected();
}

void IconListView::keyPressEvent( QKeyEvent *e )
{
 switch (e->key())
  {
    case Qt::Key_Enter:
    case Qt::Key_Return:
      if (!iconsLoaded)
       return;
      emit iconSelected();
      break;
    case Qt::Key_Left:
      goLeft();
      break;
    case Qt::Key_Right:
      goRight();
      break;
    case Qt::Key_Up:
      goUp();
      break;
    case Qt::Key_Down:
      goDown();
      break;
    case Qt::Key_PageDown:
      pageDown();
      break;
    case Qt::Key_PageUp:
      pageUp();
      break;
    case Qt::Key_Delete:
      deleteImage();
      break;
    default:
     if ( ((e->ascii()>='a') && (e->ascii()<='z')) ||
          ((e->ascii()>='A') && (e->ascii()<='Z')) ||
          ((e->ascii()>='0') && (e->ascii()<='9')) )
      {
       int newPos = iList->getNewPosByFirstCharacter(e->ascii());
       
       if (newPos>=0)
        changeSelected(newPos, true);
       else
        QApplication::beep();
      }
      break;
  }
 QScrollView::keyPressEvent(e);
}

void IconListView::contentsMouseMoveEvent( QMouseEvent *e)
{
  int x = e->pos().x();
  int y = e->pos().y();
  int icNum = iList->getIconFromPos(x, y);
  if (icNum == -1)
  {
    QToolTip::remove(viewport());
    return;
  }
  if (iconBelowMouse != icNum)
  {
   iconBelowMouse = icNum;
   setToolTip(icNum);
  }
}

void IconListView::setToolTip(int iconNum)
{
 QString toolTip;

 IconData* ic = iList->getIconData(iconNum);
 imageInfo *iI = ic->getImageInfo();
 if (ic->isDir())
  {
   ic->getNumDirsFiles();
   toolTip += QString("%1\n%2 %3\n%4 %5").arg(ic->getName())
    .arg(ApaintConfig::getText(113))
    .arg(ic->getNumDirs())
    .arg(ApaintConfig::getText(114))
    .arg(ic->getNumFiles());
  }
 else
  {    
   toolTip = ic->getName();
   if (iI->subDirID)
    toolTip += QString("\n%1").arg(QString::fromLocal8Bit(getImageSubDir(iI->subDirID)));
   toolTip += QString("\n%1").arg(ic->getInfo(true,true));
   if (ic->getXDpi() && ic->getYDpi())
    {
     if (ic->getXDpi() == ic->getYDpi())
      {
       toolTip += QString("\n%1 dpi, %2 %3/cm").arg(ic->getXDpi())
	.arg((int)((double)ic->getYDpi()/2.54+0.5))
	.arg(ApaintConfig::getText(PIXELS_MSG));
      }
     else
      {
       toolTip += QString("\n%1x%2 dpi, %3x%4 %5/cm").arg(ic->getXDpi()).arg(ic->getYDpi())
	.arg((int)((double)ic->getXDpi()/2.54+0.5)).arg((int)((double)ic->getYDpi()/2.54+0.5))
	.arg(ApaintConfig::getText(PIXELS_MSG));
      }

     toolTip += QString("\n%1 x %2 cm")
      .arg(QString().sprintf("%.1f", (ic->getImageX()*2.54)/ic->getXDpi()))
      .arg(QString().sprintf("%.1f", (ic->getImageY()*2.54)/ic->getYDpi()));

     toolTip += QString("\n%1 x %2 in")
      .arg(QString().sprintf("%.1f", (ic->getImageX()*1.0)/ic->getXDpi()))
      .arg(QString().sprintf("%.1f", (ic->getImageY()*1.0)/ic->getYDpi()));
    }

   toolTip += QString("\n%1\n%2").arg(ic->getSize()).arg(ic->getDTime());
  }
 QToolTip::remove(viewport());
 QToolTip::add(viewport(), toolTip);
}
