/***************************************************************************
                          penbrushset.h  -  description
                             -------------------
    begin                : Wed Apr 16 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PENBRUSHSET_H
#define PENBRUSHSET_H

#include <qwidget.h>
#include <qpixmap.h>
#include <qcombobox.h>
#include <qbrush.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qframe.h>
#include "toolMagicWand.h"


#define BRUSH_NUM   3
#define LINES_NUM   5

class QSpinBox;
class QLabel;
class QCheckBox;
class ApaintDoc;
class Layer;
class ArahColorDialog;

/***********************ColorChangeBtn CLASS***************************************************/

enum WhichColor {Foreground, Background};

class ColorChangeBtn : public QWidget  {
  Q_OBJECT
public:
	ColorChangeBtn(QWidget *parent, const char *name);
	~ColorChangeBtn();

  QColor foreground() const;
  QColor background() const;
  int fgIndex() {return fgNum;}
  int bgIndex() {return bgNum;}
  QSize sizeHint() const;
  void changeFgGb();
  void showColorDialog();

signals:
  void fgChanged(QRgb rgb);
  void bgChanged(QRgb rgb);
  void fgChangedIndx(int indx);
  void bgChangedIndx(int indx);
  void setColorSigIndx(int indx);
  void setColorSig(QRgb rgb);
  void setCurrentColor(QRgb rgb);

public slots:
  void setForeground(const QColor &c);
  void setBackground(const QColor &c);
  void setColorSlot(QRgb rgb, int colNum);
  void setColorSlot(QRgb rgb);
  void refresh(int maxCol);
  void colorChgSlot(QRgb c);
  void setBgFgIndxCol(int fgn, QRgb fgc, int bgn, QRgb bgc);


protected:
  void geometry(QRect &fgRect, QRect &bgRect, QRect &colRect);
  void paintEvent(QPaintEvent *e);
  void mousePressEvent(QMouseEvent *e);
  void mouseDoubleClickEvent(QMouseEvent *e);
  void setSelected(WhichColor c);

private:
  void doOnClickEvent(QMouseEvent *e, bool dclick);
  void drawNumber(QPainter *p,  QRect r, int n, WhichColor ct);

  QBrush fg, bg;
  QPixmap *arrow;
  WhichColor selected;
  int fgNum, bgNum;
  ArahColorDialog *colDlg;
  bool refreshed;
};

/************COORDS WIDGET***************/

class CoordsWidget: public QFrame {
  Q_OBJECT
public:
 CoordsWidget(QWidget *parent, const char *name);
 ~CoordsWidget();


 void setMainSize(QRect size) { mainSize = size; }
public slots:
 void showCoords(int x, int y) ;
 void showSelection(QRect size);
 void clearSelection();
 void setRGB(uchar r, uchar g, uchar b, uchar indx);
 void setSelectProp(SelectProp *sProp);
 
private:
 QLabel *lbXtext, *lbYtext, *lbx, *lby, *lbSelection;
 QHBoxLayout *hbx, *hby;
 QHBoxLayout *hbox;
 QVBoxLayout *vbox;
 QSpacerItem *sTop, *sBot;
 QString xv, yv, selSize;
 QRect mainSize;
 SelectProp *selectProp;
};  


/************PenBrushSet******************/

class PenBrushSet : public QFrame  {
  Q_OBJECT
public: 
	PenBrushSet(QWidget *parent, const char *name);
	~PenBrushSet();
//  QSize sizeHint() const;
  void initPenBrush();
  void incLineSize();
  void decLineSize();
  int getLineSize();
  void switchBrushStyle();
  
signals:
  void lineSizeChg(const QPen &p);
  void brushChg(const QBrush &b);
  
public slots:
  void brushStyleSlot();
  void lineSizeSlot(int indx);
  
private:
  void setBrushes();
  void setPenLines();
  void setBrushStyle(int i, QBrush &b);
  QSpinBox *lineSize;
  // BrushWidget *brushStyle;
  QCheckBox *brushStyle;
  QHBoxLayout *hbox;
  QVBoxLayout *vbox;
  QSpacerItem *sTop, *sBot, *sMid, *sRight;
  QBrush noB, outSolid;
};

/***********************PenBrushUtil CLASS*******************************************/

class PenBrushUtil : public QWidget {
  Q_OBJECT
public:
  PenBrushUtil(QWidget *parent, const char *name, ApaintDoc *docp);
  ~PenBrushUtil();

  uint foreground() const;
  uint background() const;
  int fgIndex() {return ccBtn->fgIndex();}
  int bgIndex() {return ccBtn->bgIndex();}
  void putInWidget(QWidget *w);
  void incLineSize();
  void decLineSize();
  int getLineSize();
  void changeFgBg();
  void switchBrush();
  void showColorDlg();
  void setSelectProp(SelectProp *sProp);

signals:
  void fgChanged(QRgb rgb);
  void bgChanged(QRgb rgb);
  void fgChangedIndx(int indx);
  void bgChangedIndx(int indx);
  void setColorSigIndx(int i);
  void setColorSig(QRgb rgb, bool doUndo);

  void sendEvents(bool on);
  void lineSizeChg(const QPen &);
  void brushChg(const QBrush&);
  void refresh(int, int, int, int);
  void colorChanged(const QColor &col, int num);
  void putColorUndo(QRgb col, int colIndx);

public slots:
  void setColorSlot(QRgb rgb);
  void setColorSlot(QRgb rgb, int i);
  void layerChangedSlot(Layer *sel);
  void initPenBrush();
  void showCoords(int x, int y);
  void showSizeSlot(QRect pos);
  void clearSizeSlot();
  void colorChgSlot(QRgb col);
  void setBgFgIndxCol(int fgn, QRgb fgc, int bgn, QRgb bgc);
  void setRGB(uchar r, uchar g, uchar b, uchar indx);

protected slots:
  void sendColorSig(QRgb rgb);

private:
   ColorChangeBtn *ccBtn;
   QVBoxLayout *bl;
   QHBoxLayout *hbl, *hbl2;
   ApaintDoc *doc;
   PenBrushSet *pbSet;
   CoordsWidget *cwid;
   QSpacerItem *sRight, *sLeft, *sMid, *sLeft2, *sRight2;
};

#endif
